/**
 * @file x-grammar.y
 * @brief Описание правил для синтаксического разбора
 * @author Маркин Алексей (alexanius.ru)
 */

%locations

%{
  /*
   * Это первая секция файла для bison. В неё можно поместить код на C,
   * необходимый для работы анализатора. Обычно здесь размещаются подключения
   * заголовочных файлов и сигнатуры функций
   */

  #include <stdio.h>
  #include <assert.h>
  #include <stdbool.h>

  #include "x-lexer.h"
  #include "x-node.h"

  /**
   * @brief Флаг наличия ошибки
   */
  static bool GotError = false;

  /**
   * @brief Флаг необходимости печати сообщения об ошибке
   */
  static bool NeedErrorMessage = false;

  /**
   * @brief Имя модуля, который разбираем
   */
  static const char * ModuleName;

  /* Это сигнатуры необходимых bison'у функций */
  int yylex (void);
  void yyerror (Node_t * moduleNode, char const * s);

  /* Здесь сигнатуры наших функций */
  static void printErrorMessage(const char * msg);
%}


/*
 * Сюда помещаем код, который будет помещаться перед всеми bison'овскими
 * объявлениями
 */
%code requires {
  #include "x-node.h"
}


/*
 * Сюда помещаем код, который понадобится в заголовочном файле
 */
%code provides
{
  #include <stdbool.h>
  #include "x-node.h"

  bool isGotError(void);
  void setModuleName(const char * name);
}

/*
 * Второй секцией идут объявления. Здесь располагаются объявления токенов и
 * типов данных, с которыми работает парсер
 */

/*
 * Данная директива указывает дополнительные аргументы для функций yyparse
 * и yyerror
 */
%parse-param {Node_t * moduleNode}

/*
 * Объявление %union содержит все возможные типы значений токенов.
 */
%union
{
   Node_t   * nodeVal;
   char     * strVal;
}

 /*
  * Ключевое слово %token служит для объявления терминала грамматики
  */
%token TOK_IF TOK_ELIF TOK_ELSE TOK_FOR TOK_BREAK TOK_CONTINUE TOK_RET TOK_DELETE
%token TOK_TYPE_INT TOK_TYPE_CHAR TOK_TYPE_BOOL TOK_TYPE_VOID
%token TOK_IDENT TOK_CHAR TOK_INT TOK_TRUE TOK_FALSE
%token TOK_LOGIC_AND TOK_LOGIC_OR TOK_IS_EQ TOK_IS_NOT_EQ
%token TOK_IS_GEQ TOK_IS_LEQ
%token TOK_ERROR

/*
 * Ключевое слово %type служит для объявления нетерминала грамматики и задания
 * его типа
 */
%type<nodeVal> input func_def func_sign sign_arg_cons sign_arg_list func_body
%type<nodeVal> statement single_statement decl_statement decl_list
%type<nodeVal> assign_expr expr literal_token call_expr arg_cons
%type<nodeVal> if_statement if_statement_head elif_statement
%type<nodeVal> for_statement loop_expr_1 loop_expr_2 loop_expr_3
%type<nodeVal> return_expr return_expr_tail delete_statement
%type<nodeVal> arr_decl_token arr_index_token
%type<strVal> ident_token

/*
 * Деструкторы вызываются автоматически если правило отменилось
 * из-за возникновения ошибки
 */
%destructor { if($$ != NULL) node_DeleteNode($$); } <nodeVal>
%destructor { free($$); } <strVal>

/*
 * Помимо директивы %token для объявления терминалов можно использовать
 * следующие директивы: %left, %right, %nonassoc, %precedence. Первые два,
 * очевидно, задают левую и правую ассоциативность соответственно. Например
 * при задании %left TOK, выражение x TOK y TOK z будет рассматриваться как
 * (x TOK y) TOK z
 *
 * %noassoc лишает токен ассоциативности. Т.е. выражение x TOK y TOK z вызовет
 * ошибку
 *
 * %precedence служит для задания приоритета терминальным символам
 */
%left '-' '+'
%left '*' '/'
%left TOK_LOGIC_AND TOK_LOGIC_OR TOK_IS_GEQ TOK_IS_LEQ '<' '>'
%left TOK_IS_EQ TOK_IS_NOT_EQ
%precedence TOK_UMIN    /* Унарный минус для отрицательных чисел */
%precedence TOK_NOT     /* Отрицание для логических значений */

/*
 * Директива %start задаёт стартовый символ грамматики
 */
%start program

%%

/*
 * Секция грамматических правил. В ней содержится описание грамматики
 * в БНФ (Бэкус-Нуровская Форма).
 *
 * Правила описываются следующим синтаксисом:
 *
 * результат: правило_1 {действия}
 *            | правило_2 {действия}
 *            ...
 *            ;
 */

/* Аксиома грамматики */
program:
  input {
      if( $1 != NULL )
      {
        /* С точки зрения грамматики функция в модуле должна быть всегда,
         * однако при обработке ошибок может получиться что мы не смогли
         * построить ни одного узла с функциями */
        node_SetModuleFirstFunc(moduleNode, $1);
      }
  }
;

/* Здесь идут входные выражения. Выражением может быть только определение функции */
input:
  func_def          { $$ = $1;}
| func_def input    {
      if( $1 != NULL )
      {
          $$ = $1;
          if( $2 != NULL )
          {
              node_SetNextNode($1, $2);
          }
      } else
      {
          $$ = $2;
      }
  }

/* Обработка ошибок */
| TOK_ERROR input   {
      /* Здесь обрабатываем случайные неизвестные символы вместо объявления функций */
      yyerror(NULL, yytext);
      printErrorMessage("недопустимые символы в программе");
      $$ = $2;
  }
| TOK_ERROR         {
      /* Здесь обрабатываем случайные неизвестные символы вместо объявления функций */
      yyerror(NULL, yytext);
      printErrorMessage("недопустимые символы в программе");
      $$ = NULL;
  }
;

/* Здесь описание и определение функции */
func_def:
  func_sign '{' func_body '}'   {
      $$ = $1;
      if( $$ != NULL )
      {
          /* Значит всё без ошибок */
          node_SetFunctionBody($$, $3);
      } else if( $3 != NULL )
      {
          /* При построении описания функции произошла ошибка,
           * и надо удалить подвисшее тело */
          node_DeleteNode($3);
      }

  }
| func_sign ';'                 {
      $$ = $1;
      if( $1 != NULL )
      {
          node_SetNodeDeclaration($1, true);
      }
  }

/* Обработка ошибок */
| error '{' func_body '}'       {
      printErrorMessage("некорректная попытка описания функции");
      node_DeleteNode($3);
      $$ = NULL;
  }
| func_sign error {
      printErrorMessage("некорректное описание тела функции");
      $$ = $1;
  }
;

/* Сигнатура функции */
func_sign:
  type_token ident_token sign_arg_cons                  {
      $$ = node_CreateFunctionNode($2);
      if( $3 != NULL )
      {
        node_SetFunctionFirstArg($$, $3);
      }
      free($2);
  }
| type_token arr_decl_token ident_token sign_arg_cons   {
      $$ = node_CreateFunctionNode($3);
      if( $4 != NULL )
      {
        node_SetFunctionFirstArg($$, $4);
      }
      free($3);
      node_SetFunctionRetNode($$, $2);
  }

/* Обработка ошибок */
| error                         {
      printErrorMessage("некорректный тип функции");
  } ident_token sign_arg_cons   {
      $$ = NULL;
      free($3);
      assert($4 == NULL); /* Пока не придумали ломающего теста */
  }

| type_token error              {
    printErrorMessage("некорректный идентификатор функции");
  } sign_arg_cons {
      $$ = NULL;
      assert($4 == NULL); /* Пока не придумали ломающего теста */
  }

| type_token ident_token error  {
      free($2);
      printErrorMessage("некорректное описание аргументов функции");
      $$ = NULL;
  }
;

/* Список аргументов */
sign_arg_cons:
  '(' ')'               { $$ = NULL; }
| '(' sign_arg_list ')' { $$ = $2; }
;

/* Непосредственно перечисление аргументов */
sign_arg_list:
  type_token obj_modifier ident_token                       {
      $$ = node_CreateObjectNode($3);
      node_SetNodeDeclaration($$, true);
      free($3);
  }
| type_token arr_decl_token ident_token                     {
      Node_t * obj = node_CreateObjectNode($3);
      free($3);
      node_SetNodeDeclaration(obj, true);
      node_SetExprNodeDecompObj($2, obj);
      $$ = $2;
  }
| type_token obj_modifier ident_token ',' sign_arg_list     {
      $$ = node_CreateObjectNode($3);
      node_SetNodeDeclaration($$, true);
      node_SetNextNode($$, $5);
      free($3);
  }
| type_token arr_decl_token ident_token ',' sign_arg_list   {
      Node_t * obj = node_CreateObjectNode($3);
      node_SetNodeDeclaration(obj, true);
      free($3);
      node_SetExprNodeDecompObj($2, obj);
      node_SetNextNode($2, $5);
      $$ = $2;
  }

/* Обработка ошибок */
| error {
      printErrorMessage("некорректное описание аргументов функции");
      $$ = NULL;
  }
;

/* Здесь у нас тело функции */
func_body:
  func_body statement   {
      $$ = $1;
      if( $$ != NULL )
      {
          node_SetNextNode(node_GetLastNode($$), $2);
      } // TODO убрать условие
      else { $$ = $2; }
  }
| %empty                { $$ = NULL; }

/* Обработка ошибок */
|  func_body error   {
      printErrorMessage("некорректное описание тела функции");
      assert($1 == NULL); /* Пока не придумали ломающего теста */
      $$ = NULL;
  }
;

/* Разбираем выражения, допустимые в теле функции */
statement:
  single_statement ';'  { $$ = $1; }
| if_statement          { $$ = $1; }
| for_statement         { $$ = $1; }

/* Обработка ошибок */
| single_statement error    {
      printErrorMessage("не хватает ';' после выражения");
      if( $1 != NULL )
      {
       node_DeleteNode($1);
      }
      $$ = NULL;
  }
;

/* Выражаение, заканчивающееся на ; */
single_statement:
  decl_statement    { $$ = $1; }
| delete_statement  { $$ = $1; }
| call_expr         { $$ = $1; }
| assign_expr       { $$ = $1; }
| return_expr       { $$ = $1; }
| TOK_BREAK         { $$ = node_CreateBreakNode();}
| TOK_CONTINUE      { $$ = node_CreateContinueNode(); }

/* Обработка ошибок */
| ident_token error {
      free($1);
      printErrorMessage("идентификатор неизвестного предназначения");
      $$ = NULL;
  }
| literal_token {
      yyerror(NULL, "");
      printErrorMessage("выражение не может начинаться с константы");
      node_DeleteNode($1);
      $$ = NULL;
  }
;

/* Здесь у нас объявление переменных */
decl_statement:
  type_token ident_token decl_list                          {
      /* Здесь имеем случай объявления обычного объекта */
      $$ = node_CreateObjectNode($2);
      node_SetNodeDeclaration($$, true);
      node_SetNextNode($$, $3);
      free($2);
  }
| type_token arr_index_token ident_token decl_list          {
      /* Здесь у нас случай объявления массива с вычислением количества его элементов */
      Node_t * obj = node_CreateObjectNode($3);
      free($3);
      node_SetNodeDeclaration(obj, true);
      node_SetExprNodeDecompObj($2, obj);
      $$ = $2;
      node_SetNextNode($$, $4);
  }
| type_token ident_token '=' expr decl_list                 {
      /* Конструкции вида int a = 1; на уровне дерева разбора
       * приводим к виду int a; a = 1; это унификация чтобы уменьшить
       * количество обрабатываемых случаев */

      Node_t * obj = node_CreateObjectNode($2);
      Node_t * decl = node_CreateObjectNode(
                        node_GetObjectName( obj));
      node_SetNodeDeclaration(decl, true);

      Node_t *  assign = node_CreateExpressionNode(EXPR_ASSIGN_T, obj, $4);

      node_SetNextNode(decl, assign);
      node_SetNextNode(assign, $5);
      $$ = decl;
      free($2);
  }
| type_token arr_decl_token ident_token '=' expr decl_list  {
      /* Конструкции вида int[] a = foo(); на уровне дерева разбора
       * приводим к виду int[] a; a = foo(); это унификация чтобы уменьшить
       * количество обрабатываемых случаев */

      Node_t * decl = node_CreateObjectNode($3);
      node_SetNodeDeclaration(decl, true);
      node_SetExprNodeDecompObj($2, decl);

      Node_t * obj = node_CreateObjectNode($3);

      /* NOTE для присваивания у нас не будет выставляться подэлемент массива,
       * т.к. по сути мы к нему не обращаемся */
      Node_t *  assign = node_CreateExpressionNode(EXPR_ASSIGN_T, obj, $5);

      node_SetNextNode($2, assign);
      node_SetNextNode(assign, $6);
      $$ = $2;
      free($3);
  }
;

/* Список объявления переменных через запятую */
decl_list:
  ',' ident_token {
      $$ = node_CreateObjectNode($2);
      node_SetNodeDeclaration($$, true);
      free($2);
      // TODO рекурсия нужна
  }
| ',' assign_expr {
      Node_t * obj = node_GetExprNodeLeftNode($2);
      Node_t * decl = node_CreateObjectNode(
                        node_GetObjectName( obj));
      node_SetNodeDeclaration(decl, true);
      node_SetNextNode(decl, $2);
      $$ = decl;
  }
| %empty          { $$ = NULL; }
;

/* Различные условные конструкции */
if_statement:
  if_statement_head                             { $$ = $1; }
| if_statement_head TOK_ELSE '{' func_body '}'  {
      if( $1 != NULL )
      {
          /* Это нормальная ситуация */
          Node_t * elseNode = node_CreateIfNode(NULL, $4);
          node_SetIfNodeElse(node_GetIfNodeLastElif($1), elseNode);
          $$ = $1;
      } else
      {
          /* Здесь обрабатываем ошибку */
          if( $4 != NULL )
          {
              node_DeleteNode($4);
          }
          $$ = NULL;
      }
  }
;

/* Общая голова для любой условной конструкции */
if_statement_head:
  TOK_IF '(' expr ')' '{' func_body '}' elif_statement {
      $$ = node_CreateIfNode($3, $6);
      if( $8 != NULL )
      {
        node_SetIfNodeElif(node_GetLastNode($$), $8);
      }
  }

/* Обработка ошибок */
| TOK_IF '(' error ')' '{' func_body '}' elif_statement {
      printErrorMessage("некорректное выражение внутри условного оператора if");
      if( $6 != NULL )
      {
        node_DeleteNode($6);
      }
      assert($8 == NULL); /* Пока не придумали ломающего теста */
      $$ = NULL;
  }
| TOK_IF error '{' func_body '}' elif_statement {
      printErrorMessage("условный оператор if требует проверки внутри круглых скобок");
      if( $4 != NULL )
      {
        node_DeleteNode($4);
      }
      assert($6 == NULL); /* Пока не придумали ломающего теста */
      $$ = NULL;
  }
| TOK_IF error {
      printErrorMessage("условный оператор if требует проверки внутри круглых скобок");
      $$ = NULL;
  }
;

/* Продолжение цепочки условных конструкций */
elif_statement:
  TOK_ELIF '(' expr ')' '{' func_body '}' elif_statement    {
      $$ = node_CreateIfNode($3, $6);
      if( $8 != NULL )
      {
        node_SetIfNodeElif(node_GetIfNodeLastElif($$), $8);
      }
  }
| %empty                                                    {
      $$ = NULL;
  }

/* Обработка ошибок */
| TOK_ELIF '(' error ')' '{' func_body '}' elif_statement {
      printErrorMessage("некорректное выражение внутри условного оператора elif");
      if( $6 != NULL )
      {
        node_DeleteNode($6);
      }
      assert($8 == NULL); /* Пока не придумали ломающего теста */
      $$ = NULL;
  }
;

/* Цикловая конструкция */
for_statement:
  TOK_FOR '(' loop_expr_1 ';' loop_expr_2 ';' loop_expr_3 ')' '{' func_body '}' {
      $$ = node_CreateForNode($3, $5, $7, $10);
  }

/* Обработка ошибок */
| TOK_FOR '(' error ';' loop_expr_2 ';' loop_expr_3 ')' '{' func_body '}' {
      printErrorMessage("некорректное выражение в секции инициализации цикла");
      node_DeleteNode($5);
      node_DeleteNode($7);
      node_DeleteNode($10);
      $$ = NULL;
  }
| TOK_FOR '(' loop_expr_1 ';' error ')' '{' func_body '}' {
      printErrorMessage("некорректное выражение в секции условия цикла");
      node_DeleteNode($3);
      node_DeleteNode($8);
      $$ = NULL;
  }
| TOK_FOR '(' loop_expr_1 ';' loop_expr_2 ';' error ')' '{' func_body '}' {
      printErrorMessage("некорректное выражение в секции индукции цикла");
      node_DeleteNode($3);
      node_DeleteNode($5);
      node_DeleteNode($10);
      $$ = NULL;
  }
| TOK_FOR '(' error ')' '{' func_body '}' {
      printErrorMessage("некорректное описание цикла");
      node_DeleteNode($6);
      $$ = NULL;
  }
| TOK_FOR error '{' func_body '}' {
      printErrorMessage("после оператора for необходимы круглые скобки с описанием");
      node_DeleteNode($4);
      $$ = NULL;
  }
;

/* Опционально заводим переменную в первой секции цикла */
loop_expr_1:
  assign_expr               { $$ = $1; }
| type_token assign_expr    {
      Node_t * obj  = node_GetExprNodeLeftNode($2);
      Node_t * decl = node_CreateObjectNode(
                        node_GetObjectName( obj));
      node_SetNodeDeclaration(decl, true);

      node_SetNextNode(decl, $2);
      $$ = decl;
  }
| %empty                    { $$ = NULL; }
;

/* Опциональное выражение во второй секции цикла (подразумевает проверку) */
loop_expr_2:
  expr      { $$ = $1; }
| %empty    { $$ = NULL; }
;

/* Опциональное выражение в третьей секции цикла (работа с индукцией) */
loop_expr_3:
  assign_expr   { $$ = $1; }
| %empty        { $$ = NULL; }
;

delete_statement:
  TOK_DELETE ident_token {
      Node_t * obj = node_CreateObjectNode($2);
      $$ = node_CreateDeleteNode(obj);
      free($2);
  }
;

/* Здесь обрабатывается вызов функции */
call_expr:
  ident_token '(' arg_cons ')'  {
      Node_t * obj = node_CreateObjectNode($1);
      free($1);
      $$ = node_CreateExpressionNode(EXPR_CALL_T, obj, $3);
  }
;

/* Выражение завершения процедуры */
return_expr:
  TOK_RET return_expr_tail {
      $$ = node_CreateReturnNode($2);
  }
;

/* Завершение процедуры может быть либо по значению, либо пустое */
return_expr_tail:
  expr      { $$ = $1; }
| %empty    { $$ = NULL; }
;

/* Выражение с присваиванием */
assign_expr:
  ident_token '=' expr                  {
      Node_t * obj = node_CreateObjectNode($1);
      free($1);
      $$ = node_CreateExpressionNode(EXPR_ASSIGN_T, obj, $3);
  }
| ident_token arr_index_token '=' expr  {
      Node_t * obj = node_CreateObjectNode($1);
      free($1);
      node_SetExprNodeDecompObj($2, obj);
      $$ = node_CreateExpressionNode(EXPR_ASSIGN_T, $2, $4);
  }
;

/* Обрабатываем последовательность аргументов */
arg_cons:
  expr                      { $$ = $1; }
| expr ',' arg_cons         { $$ = $1; node_SetNextNode($$, $3); }
| %empty                    { $$ = NULL; }
;

/* Допустимые выражения. Здесь допускается только rvalue (т.е. непосредственно вычисления) */
expr:
  literal_token                 { $$ = $1; }
| call_expr                     { $$ = $1; }
| ident_token                   {
      $$ = node_CreateObjectNode($1);
      free($1);
  }
| ident_token arr_index_token   {
      Node_t * obj = node_CreateObjectNode($1);
      free($1);
      node_SetExprNodeDecompObj($2, obj);
      $$ = $2;
  }
| '(' expr ')'                  { $$ = $2; }
| expr '+' expr                 {
      $$ = node_CreateExpressionNode(EXPR_ADD_T, $1, $3);
  }
| expr '-' expr                 {
      $$ = node_CreateExpressionNode(EXPR_SUB_T, $1, $3);
  }
| expr '*' expr                 {
      $$ = node_CreateExpressionNode(EXPR_MUL_T, $1, $3);
  }
| expr '/' expr                 {
      $$ = node_CreateExpressionNode(EXPR_DIV_T, $1, $3);
  }
| expr '<' expr                 {
      $$ = node_CreateExpressionNode(EXPR_LES_T, $1, $3);
  }
| expr '>' expr                 {
      $$ = node_CreateExpressionNode(EXPR_GRT_T, $1, $3);
  }
| expr TOK_LOGIC_AND expr       {
      $$ = node_CreateExpressionNode(EXPR_LAND_T, $1, $3);
  }
| expr TOK_LOGIC_OR expr        {
      $$ = node_CreateExpressionNode(EXPR_LOR_T, $1, $3);
  }
| expr TOK_IS_EQ expr           {
      $$ = node_CreateExpressionNode(EXPR_EQ_T, $1, $3);
  }
| expr TOK_IS_NOT_EQ expr       {
      $$ = node_CreateExpressionNode(EXPR_NEQ_T, $1, $3);
  }
| expr TOK_IS_GEQ expr          {
      $$ = node_CreateExpressionNode(EXPR_GEQ_T, $1, $3);
  }
| expr TOK_IS_LEQ expr          {
      $$ = node_CreateExpressionNode(EXPR_LEQ_T, $1, $3);
  }
| '-' expr  %prec TOK_UMIN      {
      $$ = node_CreateExpressionNode(EXPR_UMIN_T, NULL, $2);
  }
| '!' expr  %prec TOK_NOT       {
      $$ = node_CreateExpressionNode(EXPR_LNOT_T, NULL, $2);
  }
;

/* Здесь обрабатываем лексему типа */
type_token:
  TOK_TYPE_INT  { /* Типы пока что игнорируем */ }
| TOK_TYPE_CHAR { /* Типы пока что игнорируем */ }
| TOK_TYPE_BOOL { /* Типы пока что игнорируем */ }
| TOK_TYPE_VOID { /* Типы пока что игнорируем */ }
;

/* Это опциональные модификаторы объекта */
obj_modifier:
  '&'       { /* Модификатор пока что игнорируем */ }
| %empty    { /* Модификатор пока что игнорируем */ }
;


/* Обработка оператора декомпозиции с указанным индексом */
arr_index_token:
  '[' expr ']'  arr_index_token {
      Node_t * decomp = node_CreateExpressionNode(EXPR_DECOMP_T, NULL, $2);
      node_SetExprNodeDecompTarg($4, decomp);
      $$ = $4;
  }
| '[' expr ']'                  {
      $$ = node_CreateExpressionNode(EXPR_DECOMP_T, NULL, $2);
  }
;

/* Объявление массива для случаев когда размер заранее неизвестен */
arr_decl_token:
  '[' ']' arr_decl_token    {
      Node_t * decomp = node_CreateExpressionNode(EXPR_DECOMP_T, NULL, NULL);
      node_SetExprNodeDecompTarg($3, decomp);
      $$ = $3;
  }
| '[' ']'                   {
  $$ = node_CreateExpressionNode(EXPR_DECOMP_T, NULL, NULL);
  }
;

/* Здесь обрабатываем лексему идентификатора */
ident_token:
  TOK_IDENT { $$ = strdup(yytext); }
;

/* Здесь обрабатываем литеральные значения */
literal_token:
  TOK_INT       { $$ = node_CreateLiteralNode(yytext); }
| TOK_TRUE      { $$ = node_CreateLiteralNode(yytext); }
| TOK_FALSE     { $$ = node_CreateLiteralNode(yytext); }
| TOK_CHAR      { $$ = node_CreateLiteralNode(yytext); }
;

%%

/**
 * @brief Реакция на ошибку
 *
 * @param [in] msg Печатаемое сообщение
 *
 * Данная функция автоматически вызывается bison'ом если на вход поступает
 * токен, не удовлетворяющий ни одному правилу
 */
void yyerror(Node_t * moduleNode, char const * msg)
{
    assert(!NeedErrorMessage);
    fprintf (stderr,
             "%s:%d:%d: '%s' ",
             ModuleName,
             yylloc.first_line,
             yylloc.first_column,
             yytext);
    GotError            = true;
    NeedErrorMessage    = true;
}

/**
 * @brief Печатаем остаток сообщения об ошибке если необходимо
 *
 * @param [in] msg  Печатаемое сообщение
 */
static void printErrorMessage(const char * msg)
{
    if( !NeedErrorMessage )
    {
        return;
    }

    fprintf(stderr, "%s\n", msg);
    NeedErrorMessage = false;
}

/**
 * @brief Возвращаем значение флага наличия ошибки
 *
 * @returns Значение флага наличия ошибки
 */
bool isGotError(void)
{
    return GotError;
}

/**
 * @brief Задаём имя текущего модуля
 *
 * @param [in] name  Имя модуля
 */
void setModuleName(const char * name)
{
    ModuleName = name;
}
