/**
 * @file x-lexer.l
 * @brief Описание правил для лексического разбора
 * @author Маркин Алексей (alexanius.ru)
 */

/* Эти три директивы говорят flex не генерировать функции input, yyunput и
 * yywrap, которые в данном случае не используются. Они необходимы для
 * подавления предупреждений компилятора */
%option nounput
%option noinput
%option noyywrap

%{

#include <stdio.h>
#include <stdbool.h>
#include <assert.h>

#include "x-grammar.h"

/**
 * @brief Контейнер для положения символа в тексте
 */
typedef struct
{
    int line;   /*!< @brief Строка */
    int colomn; /*!< @brief Столбец */
} Position_t;

/**
 * @brief Переменная, в которой храним текущую позицию
 */
static Position_t Position = {1, 1};

/**
 * @brief Сдвигаемся на #num строк
 *
 * @param [in] num  Количество строк, на которое сдвигаемся
 */
static void incrLine(int num) {Position.line += num;}

/**
 * @brief Получаем текущую строку в читаемом файле
 *
 * @returns Текущая строка в читаемом файле
 */
static int line(void) {return Position.line;}

/**
 * @brief Сдвигаемся на #num столбцов
 *
 * @param [in] num  Количество столбцов, на которое сдвигаемся
 */
static void incrColomn(int num) {Position.colomn += num;}

/**
 * @brief Получаем текущий столбец в читаемом файле
 *
 * @returns Текущий столбец в читаемом файле
 */
static int colomn(void) {return Position.colomn;}

/**
 * @brief Сбрасываем текущий столбец в читаемом файле в начало строки
 */
static void dropColomn(void) {Position.colomn = 1;}

/**
 * @brief Автоматическое действие при чтении лексемы
 */
#define YY_USER_ACTION { \
    yylloc.first_line   = line(); \
    yylloc.last_line    = line(); \
    yylloc.first_column = colomn(); \
    yylloc.last_column  = colomn() + yyleng - 1; \
    incrColomn(yyleng); }

%}

/* Здесь описываем состояния, в которые попадает лексический анализатор */

/* Состояние многострочного коментария */
%x COMMENT

INTEGER [0-9]+
SPACE [ \t]*
IDENTIFIER [a-zA-Z_][a-zA-Z_0-9]*
NOT_AVIABLE [^a-zA-Z_0-9+\-*/><!;,=(){}\[\]&|'" \n\t]

%%

 /* Обработка комментариев */
"/*"                    { /* Включаем состояние анализа комментариев */ BEGIN(COMMENT); }
<COMMENT>.              { /* Внутри комментария ничего не делаем */ }
<COMMENT>"*/"           { /* Возвращаемся в начальное состояние */ BEGIN(INITIAL); }

 /* Ключевые слова */
if                      { return TOK_IF; }
elif                    { return TOK_ELIF; }
else                    { return TOK_ELSE; }
for                     { return TOK_FOR; }
break                   { return TOK_BREAK; }
continue                { return TOK_CONTINUE; }
return                  { return TOK_RET; }
delete                  { return TOK_DELETE; }
true                    { return TOK_TRUE; }
false                   { return TOK_FALSE; }

int                     { return TOK_TYPE_INT; }
char                    { return TOK_TYPE_CHAR; }
bool                    { return TOK_TYPE_BOOL; }
void                    { return TOK_TYPE_VOID; }

 /* Обработка единичных символов */
[+\-*/><!;,=(){}\[\]&]   { return yytext[0]; }

 /* Операции из нескольких символов */
"&&"                    { return TOK_LOGIC_AND; }
"||"                    { return TOK_LOGIC_OR; }
"=="                    { return TOK_IS_EQ; }
"!="                    { return TOK_IS_NOT_EQ; }
">="                    { return TOK_IS_GEQ; }
"<="                    { return TOK_IS_LEQ; }

 /* Обработка нормального хода программы */
\'\\?.\'                { return TOK_CHAR; }
{INTEGER}               { return TOK_INT; }
{IDENTIFIER}            { return TOK_IDENT; }
{SPACE}                 { /* Пробелы игнорируем*/ }

 /* В любом состоянии лексического анализатора учитываем переход на новую строку */
<*>\n                   { dropColomn(); incrLine(1);yylloc.first_line++;}

 /* Все недопустимые символы являются ошибкой */
{NOT_AVIABLE}+          { return TOK_ERROR; }

 /* Здесь ловим то что не удалось поймать предыдущими правилами */
.                       { return TOK_ERROR; }

%%
