/**
 * @file x-node.c
 * @brief Определение узла дерева разбора
 * @author Маркин Алексей (alexanius@gmail.com)
 */

#include "x-node.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/**
 * @brief Свойства выражений
 */
typedef struct
{
    const char  * name; /*!< Символьная запись выражения */
    int         nthy;   /*!< Арность выражения */
} ExprProperties_t;

static Node_t * node_GetModuleFirstFunc(const Node_t *moduleNode);
static ExprType_t node_GetExpressionType(const Node_t  * exprNode);

/**
 * @brief Таблица свойств выражений
 */
static ExprProperties_t ExprProperties[EXPR_END] =
{
    [EXPR_ASSIGN_T]   = {"=",   2},
    [EXPR_ADD_T]      = {"+",   2},
    [EXPR_SUB_T]      = {"-",   2},
    [EXPR_MUL_T]      = {"*",   2},
    [EXPR_DIV_T]      = {"/",   2},
    [EXPR_GRT_T]      = {">",   2},
    [EXPR_LES_T]      = {"<",   2},
    [EXPR_GEQ_T]      = {">=",  2},
    [EXPR_LEQ_T]      = {"<=",  2},
    [EXPR_EQ_T]       = {"==",  2},
    [EXPR_NEQ_T]      = {"!=",  2},
    [EXPR_LOR_T]      = {"||",  2},
    [EXPR_LAND_T]     = {"&&",  2},
    [EXPR_LNOT_T]     = {"!",   1},
    [EXPR_UMIN_T]     = {"-",   1},
    [EXPR_DECOMP_T]   = {"[]",  2},
    [EXPR_CALL_T]     = {"()",  2}
};

/**
 * @brief Таблица с символьными именами типов узлов
 */
static const char * NodeTypeToString[NODE_END] =
{
    [NODE_MODULE_T]     = "NODE_MODULE_T",
    [NODE_FUNCTION_T]   = "NODE_FUNCTION_T",
    [NODE_OBJECT_T]     = "NODE_OBJECT_T",
    [NODE_EXPRESSION_T] = "NODE_EXPRESSION_T",
    [NODE_LITERAL_T]    = "NODE_LITERAL_T",
    [NODE_IF_T]         = "NODE_IF_T",
    [NODE_FOR_T]        = "NODE_FOR_T",
    [NODE_CONTINUE_T]   = "NODE_CONTINUE_T",
    [NODE_BREAK_T]      = "NODE_BREAK_T",
    [NODE_RETURN_T]     = "NODE_RETURN_T",
    [NODE_DELETE_T]     = "NODE_DELETE_T"
};

/**
 * @brief Создаём узел заданного типа
 *
 * @param [in] type  Тип создаваемого узла
 *
 * @returns Созданный узел
 */
static Node_t * node_CreateNode(NodeType_t  type)
{
    Node_t  * node; /* Создаваемый узел */

    node = calloc(1, sizeof(Node_t));
    assert(node != NULL);

    node->type = type;

    return node;
}

/**
 * @brief Получаем тип узла
 *
 * @param [in] node  Узел, параметр которого получаем
 *
 * @return Тип узла
 */
static NodeType_t node_GetNodeType(const Node_t * node)
{
    return node->type;
}

/**
 * @brief Удаляем узел, всех его детей и всех соседей
 *
 * @param [in,out] node  Удаляемый узел
 */
void node_DeleteNode(Node_t * node)
{
    switch(node_GetNodeType(node))
    {
      case NODE_MODULE_T:
      {
        free(node->moduleName);

        Node_t * func = node_GetModuleFirstFunc(node);
        if( func != NULL)
        {
            /* Удаляем все функции модуля */
            node_DeleteNode(func);
        }
        break;
      }
      case NODE_FUNCTION_T:
      {
        free(node->funcName);

        Node_t * arg = node_GetFunctionFirstArg(node);
        if( arg != NULL)
        {
            /* Удаляем аргументы функции */
            node_DeleteNode(arg);
        }

        Node_t * retNode = node_GetFunctionRetNode(node);
        if( retNode != NULL)
        {
            /* Удаляем возвращаемое значение */
            node_DeleteNode(retNode);
        }

        Node_t * body = node_GetFunctionBody(node);
        if( body != NULL)
        {
            /* Удаляем аргументы функции */
            node_DeleteNode(body);
        }

        break;
      }
      case NODE_OBJECT_T:
      {
        free(node->objName);
        break;
      }
      case NODE_EXPRESSION_T:
      {
        Node_t * arg = node_GetExprNodeLeftNode(node);
        if( arg != NULL )
        {
           node_DeleteNode(arg);
        }

        arg = node_GetExprNodeRightNode(node);
        if( arg != NULL )
        {
            node_DeleteNode(arg);
        }

        break;
      }
      case NODE_LITERAL_T:
      {
        free(node->literalLexeme);
        break;
      }
      case NODE_IF_T:
      {
        Node_t * cond = node_GetIfNodeCondExpr(node);
        if( cond != NULL )
        {
            node_DeleteNode(cond);
        }

        Node_t * body = node_GetIfNodeBody(node);
        if( body != NULL )
        {
            node_DeleteNode(body);
        }

        Node_t * alternate = node_GetIfNodeAlternate(node);
        if( alternate != NULL )
        {
            node_DeleteNode(alternate);
        }

        break;
      }
      case NODE_FOR_T:
      {
        Node_t * init = node_GetForNodeInit(node);
        if( init != NULL )
        {
            node_DeleteNode(init);
        }

        Node_t * cond = node_GetForNodeCond(node);
        if( cond != NULL )
        {
            node_DeleteNode(cond);
        }
        Node_t * iter = node_GetForNodeIter(node);
        if( iter != NULL )
        {
            node_DeleteNode(iter);
        }

        Node_t * body = node_GetForNodeBody(node);
        if( body != NULL )
        {
            node_DeleteNode(body);
        }
        break;
      }
      case NODE_CONTINUE_T:
      case NODE_BREAK_T:
      {
        break;
      }
      case NODE_RETURN_T:
      {
        Node_t * retExpr = node_GetReturnNodeExpr(node);
        if( retExpr != NULL )
        {
            node_DeleteNode(retExpr);
        }
        break;
      }
      case NODE_DELETE_T:
      {
        node_DeleteNode(node_GetDeleteNodeObj(node));
        break;
      }
// LCOV_EXCL_START
      default:
      {
        assert(!"Неизвестный тип удаляемого узла");
      }
// LCOV_EXCL_STOP
    }

    /* Запускаем удаление следующих узлов */
    Node_t  * nextNode = node_GetNextNode(node);
    if( nextNode != NULL )
    {
        node_DeleteNode(nextNode);
    }

    free(node);
}

/**
 * @brief Печатаем узел и всех его потомков
 *
 * @param [in] node   Печатаемый узел
 * @param [in] depth  Глубина рекурсивного спуска
 */
void node_PrintNodeRec(const Node_t * node, int depth)
{
    printf("%*s", depth * 2, " "); /* Печатаем нужное количество пробелов */

    NodeType_t  nodeType = node_GetNodeType(node);
    printf("<%s: ", NodeTypeToString[nodeType]);

    switch(node_GetNodeType(node))
    {
      case NODE_MODULE_T:
      {
        printf("\"%s\">\n",
               node_GetModuleName(node));

        const Node_t * func = node_GetModuleFirstFunc(node);
        if( func != NULL)
        {
            /* Печатаем все функции модуля */
            node_PrintNodeRec(func, depth + 1);
        }
        break;
      }
      case NODE_FUNCTION_T:
      {
        printf("\"%s\"",
               node_GetFunctionName(node));

        if( node_IsDeclaration(node) )
        {
            printf(" [объявление]");
        }

        printf(">\n");

        Node_t * retNode = node_GetFunctionRetNode(node);
        if( retNode != NULL )
        {
            printf("Возвращаемый массив:\n");
            node_PrintNodeRec(retNode, depth + 1);
        }

        printf("%*s Аргументы функции:\n", depth * 2, " ");
        Node_t * arg = node_GetFunctionFirstArg(node);
        if( arg != NULL)
        {
            /* Печатаем аргументы функции */
            node_PrintNodeRec(arg, depth + 1);
        } else
        {
            printf("%*s  Отсутствуют\n", depth * 2, " ");
        }

        printf("%*s Тело функции:\n", depth * 2, " ");
        Node_t * body = node_GetFunctionBody(node);
        if( body != NULL)
        {
            /* Печатаем аргументы функции */
            node_PrintNodeRec(body, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        break;
      }
      case NODE_OBJECT_T:
      {
        printf("\"%s\"",
               node_GetObjectName(node));

        if( node_IsDeclaration(node) )
        {
            printf(" [объявление]");
        }

        printf(">\n");
        break;
      }
      case NODE_EXPRESSION_T:
      {
        printf("\"%s\"",
               ExprProperties[node_GetExpressionType(node)].name);
        printf(">\n");

        printf("%*s Левый аргумент:\n", depth * 2, " ");
        Node_t * arg = node_GetExprNodeLeftNode(node);
        if( arg != NULL )
        {
            node_PrintNodeRec(arg, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        printf("%*s Правый аргумент:\n", depth * 2, " ");
        arg = node_GetExprNodeRightNode(node);
        if( arg != NULL )
        {
            node_PrintNodeRec(arg, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        break;
      }
      case NODE_LITERAL_T:
      {
        printf("\"%s\">\n",
              node_GetLiteralLexeme(node));

        break;
      }
      case NODE_IF_T:
      {
        printf(">\n");

        Node_t * condNode = node_GetIfNodeCondExpr(node);
        printf("%*s Условие ветвления:\n", depth * 2, " ");
        if( condNode != NULL )
        {
            node_PrintNodeRec(condNode, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        Node_t * bodyNode = node_GetIfNodeBody(node);
        printf("%*s Тело ветвления:\n", depth * 2, " ");
        if( bodyNode != NULL )
        {
            node_PrintNodeRec(bodyNode, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        Node_t * alternateNode = node_GetIfNodeAlternate(node);
        printf("%*s Узел альтернативы:\n", depth * 2, " ");
        if( alternateNode != NULL )
        {
            node_PrintNodeRec(alternateNode, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        break;
      }
      case NODE_FOR_T:
      {
        printf(">\n");

        Node_t * initNode = node_GetForNodeInit(node);
        printf("%*s Узел инициализации цикла:\n", depth * 2, " ");
        if( initNode != NULL )
        {
            node_PrintNodeRec(initNode, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        Node_t * condNode = node_GetForNodeCond(node);
        printf("%*s Узел условия выхода из цикла:\n", depth * 2, " ");
        if( condNode != NULL )
        {
            node_PrintNodeRec(condNode, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        Node_t * iterNode = node_GetForNodeIter(node);
        printf("%*s Узел действия при переходе на итерацию:\n", depth * 2, " ");
        if( iterNode != NULL )
        {
            node_PrintNodeRec(iterNode, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        Node_t * bodyNode = node_GetForNodeBody(node);
        printf("%*s Узел тела цикла:\n", depth * 2, " ");
        if( bodyNode != NULL )
        {
            node_PrintNodeRec(bodyNode, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }

        break;
      }
      case NODE_CONTINUE_T:
      case NODE_BREAK_T:
      {
        printf(">\n");
        break;
      }
      case NODE_RETURN_T:
      {
        printf(">\n");

        Node_t * retExpr = node_GetReturnNodeExpr(node);
        printf("%*s Возвращаемое выражение:\n", depth * 2, " ");
        if( retExpr != NULL )
        {
            node_PrintNodeRec(retExpr, depth + 1);
        } else
        {
            printf("%*s  Отсутствует\n", depth * 2, " ");
        }
        break;
      }
      case NODE_DELETE_T:
      {
        printf(">\n");
        node_PrintNodeRec(node_GetDeleteNodeObj(node), depth + 1);
        break;
      }
// LCOV_EXCL_START
      default:
      {
        assert(!"Неизвестный тип печатаемого узла");
      }
// LCOV_EXCL_STOP
    }

    /* Запускаем печать следующих узлов */
    Node_t  * nextNode = node_GetNextNode(node);
    if( nextNode != NULL )
    {
        node_PrintNodeRec(nextNode, depth);
    }
}

/**
 * @brief Печатаем узел и всех его потомков
 *
 * @param [in] node  Печатаемый узел
 */
void node_PrintNode(const Node_t * node)
{
    node_PrintNodeRec(node, 0);
}


/**
 * @brief Получаем следующий узел
 *
 * @param [in]  Узел, параметр которого получаем
 *
 * @returns Следующий узел
 */
Node_t * node_GetNextNode(const Node_t * node)
{
    return node->nextNode;
}

/**
 * @brief Получаем последний узел цепочки
 *
 * @param [in] node  Узел, от которого ищем последний
 *
 * @returns Последний узел цепочки
 */
Node_t * node_GetLastNode(const Node_t * node)
{
    while(node->nextNode != NULL) { node = node->nextNode; }
    return (Node_t *)node;
}

/**
 * @brief Добавляем узел, следующий за данным
 *
 * @param [in,out]  node      Узел, к которому добавляем
 * @param [in]      nextNode  Добавляемый узел
 */
void node_SetNextNode(Node_t * node, Node_t * nextNode)
{
    assert(node->nextNode == NULL);
    node->nextNode = nextNode;
}

/**
 * @brief Получаем признак объявления для узла
 *
 * @param [in] node  Узел, параметр которого получаем
 *
 * @returns #true если узел является объявлением, #false в противном случае
 */
bool node_IsDeclaration(const Node_t * node)
{
    /* Такой признак имеет смысл не у всех узлов */
    assert(node_GetNodeType(node)   == NODE_FUNCTION_T ||
           node_GetNodeType(node)   == NODE_OBJECT_T ||
           /* Последнее условие нужно для успешного прохождения других проверок */
           node_GetNodeType(node)   == NODE_EXPRESSION_T );

    return node->isDecl;
}

/**
 * @brief Задаём узлу признак объявления
 *
 * @param [in,out] node    Узел, параметр которого меняем
 * @param [in]     isDecl  Признак объявления
 */
void node_SetNodeDeclaration(Node_t * node, bool isDecl)
{
    /* Такой признак имеет смысл не у всех узлов */
    assert(node_GetNodeType(node)   == NODE_FUNCTION_T ||
           node_GetNodeType(node)   == NODE_OBJECT_T);

    node->isDecl = isDecl;
}

/**
 * @brief Создаём узел модуля
 *
 * @param [in] moduleName  название модуля
 *
 * @returns Созданный узел модуля
 */
Node_t * node_CreateModuleNode(const char * moduleName)
{
    Node_t  * node;

    node = node_CreateNode(NODE_MODULE_T);

    node->moduleName    = strdup(moduleName);

    return node;
}

/**
 * @brief Поулчаем имя модуля
 *
 * @param [in] node  Узел, параметр которого получаем
 *
 * @returns Имя модуля
 */
const char * node_GetModuleName(const Node_t * node)
{
    assert(node_GetNodeType(node) == NODE_MODULE_T);
    return node->moduleName;
}

/**
 * @brief Задаём первую функцию модуля
 *
 * @param [in,out] moduleNode  Узел модуля, в который пишем
 * @param [in,out] funcNode    Узел процедуры, которую пишем в модуль
 */
void node_SetModuleFirstFunc(Node_t * moduleNode,
                             Node_t * funcNode)
{
    assert(node_GetNodeType(moduleNode) == NODE_MODULE_T);
    assert(moduleNode->firstFunc == NULL);
    assert(node_GetNodeType(funcNode)   == NODE_FUNCTION_T);

    moduleNode->firstFunc = funcNode;
}

/**
 * @brief Получаем первую функцию модуля
 *
 * @param [in] moduleNode  Узел модуля, функцию которого получаем
 *
 * @returns Первая функция модуля или NULL если таковые отсутствуют
 */
static Node_t * node_GetModuleFirstFunc(const Node_t * moduleNode)
{
    assert(node_GetNodeType(moduleNode) == NODE_MODULE_T);

    return moduleNode->firstFunc;
}

/**
 * @brief Создаём узел функции
 *
 * @param [in] functionName  Имя функции
 *
 * @returns Узел функции
 */
Node_t * node_CreateFunctionNode(const char * functionName)
{
    Node_t  * node;

    node = node_CreateNode(NODE_FUNCTION_T);

    node->funcName    = strdup(functionName);

    return node;
}

/**
 * @brief Поулчаем имя функции
 *
 * @param [in] node  Узел, параметр которого получаем
 *
 * @returns Имя функции
 */
const char * node_GetFunctionName(const Node_t * funcNode)
{
    assert(node_GetNodeType(funcNode) == NODE_FUNCTION_T);
    return funcNode->funcName;
}

/**
 * @brief Задаём первый аргумент функции
 *
 * @param [in,out] funcNode  Узел, параметр которого задаём
 * @param [in]     objNode   Узел объекта, который записываем аргументом
 */
void node_SetFunctionFirstArg(Node_t * funcNode, Node_t * objNode)
{
    assert(node_GetNodeType(funcNode) == NODE_FUNCTION_T);
    assert(node_GetFunctionFirstArg(funcNode) == NULL);
    assert(node_GetNodeType(objNode) == NODE_OBJECT_T ||
           (node_GetNodeType(objNode) == NODE_EXPRESSION_T &&
            node_GetExpressionType(objNode) == EXPR_DECOMP_T));
    assert(node_IsDeclaration(objNode) ||
           (node_GetNodeType(objNode) == NODE_EXPRESSION_T &&
            node_GetExpressionType(objNode) == EXPR_DECOMP_T));

    funcNode->firstArg = objNode;
}

/**
 * @brief Получаем первый аргумент функции
 *
 * @param [in] funcNode  Узел, параметр которого получаем
 *
 * @returns Первый аргумент функции
 */
Node_t * node_GetFunctionFirstArg(const Node_t * funcNode)
{
    assert(node_GetNodeType(funcNode) == NODE_FUNCTION_T);

    return funcNode->firstArg;
}

/**
 * @brief Задаём узел с типом возвращаемого значения
 *
 * @param [in,out] funcNode  Узел, параметр которого задаём
 * @param [in]     retNode   Узел типа возвращаемого значения
 *
 * @note Задавать узел типа возврата для нашей грамматики возможно только
 *       в случаях когда она возвращает массив
 */
void node_SetFunctionRetNode(Node_t * funcNode, Node_t * retNode)
{
    assert(node_GetNodeType(funcNode) == NODE_FUNCTION_T);
    assert(node_GetFunctionRetNode(funcNode) == NULL);
    assert(node_GetNodeType(retNode) == NODE_EXPRESSION_T &&
           node_GetExpressionType(retNode) == EXPR_DECOMP_T);

    funcNode->retTypeNode = retNode;
}

/**
 * @brief Получаем узел с типом возвращаемого значения
 *
 * @param [in] funcNode  Узел, параметр которого получаем
 *
 * @returns Узел с типом возвращаемого значения
 */
Node_t * node_GetFunctionRetNode(const Node_t * funcNode)
{
    assert(node_GetNodeType(funcNode) == NODE_FUNCTION_T);

    return funcNode->retTypeNode;
}

/**
 * @brief Задаём тело функции
 *
 * @param [in,out] funcNode  Узел, параметр которого задаём
 * @param [in]     node      Узел тела функции
 */
void node_SetFunctionBody(Node_t * funcNode, Node_t * node)
{
    assert(node_GetNodeType(funcNode) == NODE_FUNCTION_T);
    assert(node_GetFunctionBody(funcNode) == NULL);

    funcNode->funcBody = node;
}

/**
 * @brief Получаем тело функции
 *
 * @param [in] funcNode  Узел, параметр которого получаем
 *
 * @returns Тело функции
 */
Node_t * node_GetFunctionBody(const Node_t * funcNode)
{
    assert(node_GetNodeType(funcNode) == NODE_FUNCTION_T);

    return funcNode->funcBody;
}

/**
 * @brief Создаём узел объекта
 *
 * @param [in] objectName  Имя объекта
 *
 * @returns Узел объекта
 */
Node_t * node_CreateObjectNode(const char * objectName)
{
    Node_t  * node;

    node = node_CreateNode(NODE_OBJECT_T);

    node->objName    = strdup(objectName);

    return node;
}

/**
 * @brief Поулчаем имя объекта
 *
 * @param [in] node  Узел, параметр которого получаем
 *
 * @returns Имя объекта
 */
const char * node_GetObjectName(const Node_t * objNode)
{
    assert(node_GetNodeType(objNode) == NODE_OBJECT_T);
    return objNode->objName;
}


/**
 * @brief Создаём узел выражения
 *
 * @param [in] exprType  Тип выражения
 * @param [in] leftNode   Левая часть выражения
 * @param [in] rightNode  Правая часть выражения
 *
 * @returns Узел выражения
 */
Node_t * node_CreateExpressionNode(ExprType_t   exprType,
                                   Node_t       * leftNode,
                                   Node_t       * rightNode)
{
    Node_t  * node = node_CreateNode(NODE_EXPRESSION_T);
    node->exprType  = exprType;

    int nthy    = ExprProperties[exprType].nthy;

    /* Здесь присваиваем аргументы */
    if( nthy == 1 )
    {
        /* Для выражений с одним аргументом используем только правый */
        assert(leftNode == NULL);
        assert(rightNode != NULL);
        node->rightArg = rightNode;
    } else if( nthy == 2 )
    {
        /* Задаём оба аргумента выражения */
        if( exprType != EXPR_DECOMP_T )
        {
            assert(leftNode != NULL);
        }

        if( exprType != EXPR_CALL_T &&
            exprType != EXPR_DECOMP_T )
        {
            /* Для вызовов правый аргумент опционален */
            assert(rightNode != NULL);
        }
        node->leftArg = leftNode;
        node->rightArg = rightNode;
    } else
    {
// LCOV_EXCL_START
        assert(!"Неизвестная арность выражения");
// LCOV_EXCL_STOP
    }

    switch(exprType)
    {
      case EXPR_ASSIGN_T:
      {
        /* Левой частью присваивания обязан быть узел объекта,
         * а правой - объектом, выражением или константой.
         * Оба узла не могут быть объявлением */
        assert(node_GetNodeType(leftNode) == NODE_OBJECT_T ||
               (node_GetNodeType(leftNode) == NODE_EXPRESSION_T &&
                node_GetExpressionType(leftNode) == EXPR_DECOMP_T));
        assert(!node_IsDeclaration(leftNode));
        assert((node_GetNodeType(rightNode) == NODE_OBJECT_T &&
                !node_IsDeclaration(rightNode)) ||
               node_GetNodeType(rightNode) == NODE_EXPRESSION_T ||
               node_GetNodeType(rightNode) == NODE_LITERAL_T);
        break;
      }
      case EXPR_ADD_T:
      case EXPR_SUB_T:
      case EXPR_MUL_T:
      case EXPR_DIV_T:
      case EXPR_GRT_T:
      case EXPR_LES_T:
      case EXPR_GEQ_T:
      case EXPR_LEQ_T:
      case EXPR_EQ_T:
      case EXPR_NEQ_T:
      case EXPR_LOR_T:
      case EXPR_LAND_T:
      {
        /* В арифметических и логических выражениях оба аргумента могут
         * быть выражения, объекты, константы */
        assert((node_GetNodeType(leftNode) == NODE_OBJECT_T &&
                !node_IsDeclaration(leftNode)) ||
               node_GetNodeType(leftNode) == NODE_EXPRESSION_T ||
               node_GetNodeType(leftNode) == NODE_LITERAL_T);
        assert((node_GetNodeType(rightNode) == NODE_OBJECT_T &&
                !node_IsDeclaration(rightNode)) ||
               node_GetNodeType(rightNode) == NODE_EXPRESSION_T ||
               node_GetNodeType(rightNode) == NODE_LITERAL_T);
        break;
      }
      case EXPR_LNOT_T:
      case EXPR_UMIN_T:
      {
        assert(node_GetNodeType(rightNode) == NODE_OBJECT_T ||
               node_GetNodeType(rightNode) == NODE_EXPRESSION_T ||
               node_GetNodeType(rightNode) == NODE_LITERAL_T);
        break;
      }
      case EXPR_DECOMP_T:
      {
        /* NOTE пока что таких случаев не было */
        assert(leftNode == NULL);
#if 0
        if( leftNode != NULL )
        {
            assert((node_GetNodeType(leftNode) == NODE_EXPRESSION_T &&
                    node_GetExpressionType(leftNode) == EXPR_DECOMP_T) ||
                   node_GetNodeType(leftNode) == NODE_OBJECT_T);
        }
#endif /* 0 */

        if( rightNode != NULL )
        {
            assert(node_GetNodeType(rightNode) == NODE_EXPRESSION_T ||
                   node_GetNodeType(rightNode) == NODE_OBJECT_T ||
                   node_GetNodeType(rightNode) == NODE_LITERAL_T );
        }
        break;
      }
      case EXPR_CALL_T:
      {
        assert(node_GetNodeType(leftNode) == NODE_OBJECT_T);
        break;
      }
// LCOV_EXCL_START
      default:
      {
        assert(!"Неизвестный тип выражения");
        break;
      }
// LCOV_EXCL_STOP
    }

    return node;
}

/**
 * @brief Получаем тип выражения
 *
 * @param [in] exprNode  Узел, параметр которого получаем
 *
 * @returns Тип выражения
 */
static ExprType_t node_GetExpressionType(const Node_t  * exprNode)
{
    assert(node_GetNodeType(exprNode) == NODE_EXPRESSION_T);

    return exprNode->exprType;
}

/**
 * @brief Получаем левый аргумент выражения
 *
 * @param [in] exprType  Узел, параметр которого получаем
 *
 * @returns Левый аргумент выражения
 */
Node_t * node_GetExprNodeLeftNode(const Node_t * exprNode)
{
    assert(node_GetNodeType(exprNode) == NODE_EXPRESSION_T);

    return exprNode->leftArg;
}

/**
 * @brief Получаем правый аргумент выражения
 *
 * @param [in] exprType  Узел, параметр которого получаем
 *
 * @returns Правый аргумент выражения
 */
Node_t * node_GetExprNodeRightNode(const Node_t * exprNode)
{
    assert(node_GetNodeType(exprNode) == NODE_EXPRESSION_T);

    return exprNode->rightArg;
}

/**
 * @brief Устанавливаем узлу декомпозиции цель декомпозиции
 *
 * @param [in,out] decompNode  Узел, параметр которого задаём
 * @param [in]     targNode    Узел, являющийся целью декомпозиции
 *
 * Целью декомпозиции может быть либо узел с предыдущей декомпозицией,
 * либо узел объекта, являющегося массивом
 */
void node_SetExprNodeDecompTarg(Node_t * decompNode,
                                Node_t * targNode)
{
    assert(node_GetNodeType(decompNode) == NODE_EXPRESSION_T);
    assert(node_GetExpressionType(decompNode) == EXPR_DECOMP_T);
    assert((node_GetNodeType(targNode) == NODE_EXPRESSION_T &&
            node_GetExpressionType(targNode) == EXPR_DECOMP_T) ||
           node_GetNodeType(targNode) == NODE_OBJECT_T);

    decompNode->leftArg = targNode;
}

/**
 * @brief Задаём объект массива, который декомпозируем
 *
 * @param [in,out] decompNode  Узел последней декомпозиции
 * @param [in]     objNode     Объект, который будем декомпозировать
 */
void node_SetExprNodeDecompObj(Node_t * decompNode,
                               Node_t * objNode)
{
    assert(node_GetNodeType(decompNode) == NODE_EXPRESSION_T);
    assert(node_GetExpressionType(decompNode) == EXPR_DECOMP_T);
    assert(node_GetNodeType(objNode) == NODE_OBJECT_T);

    /* Спускаемся до самого нижнего узла декомпозиции */
    for( Node_t  * prevDecomp = node_GetExprNodeLeftNode(decompNode);
         prevDecomp != NULL;
         prevDecomp = node_GetExprNodeLeftNode(decompNode) )
    {
        decompNode = prevDecomp;
    }

    node_SetExprNodeDecompTarg(decompNode, objNode);
}


/**
 * @brief Создаём узел литерала
 *
 * @param [in] literalLexeme  Символьное значение литерала
 *
 * @returns Узел литерала
 */
Node_t * node_CreateLiteralNode(const char * literalLexeme)
{
    Node_t  * node;

    node = node_CreateNode(NODE_LITERAL_T);

    node->literalLexeme  = strdup(literalLexeme);

    return node;
}

/**
 * @brief Получаем лексему литерала
 *
 * @param [in] literalNode  Узел, параметр которого получаем
 *
 * @returns Лексема литерала
 */
const char * node_GetLiteralLexeme(const Node_t * literalNode)
{
    assert(node_GetNodeType(literalNode) == NODE_LITERAL_T);

    return literalNode->literalLexeme;
}

/**
 * @brief Создаём узел условия
 *
 * @param [in] condExpr  Выражение для вычисления условия
 * @param [in] bodyNode  Узел с телом по истинному переходу
 *
 * @returns Узел условия
 */
Node_t * node_CreateIfNode(Node_t * condExpr, Node_t * bodyNode)
{
    Node_t  * node;

    node = node_CreateNode(NODE_IF_T);

    if( condExpr != NULL )
    {
        /* Условие может быть нулевым если создаём else узел */
        assert(node_GetNodeType(condExpr) == NODE_EXPRESSION_T ||
               node_GetNodeType(condExpr) == NODE_LITERAL_T ||
               node_GetNodeType(condExpr) == NODE_OBJECT_T);
    }

    node->condExpr = condExpr;
    node->ifBodyNode = bodyNode;

    return node;
}

/**
 * @brief Задаём условному узлу альтернативный elif узел
 *
 * @param [in,out] ifNode    Условный узел
 * @param [in]     elifNode  Узел альтернативы
 */
void node_SetIfNodeElif(Node_t * ifNode, Node_t * elifNode)
{
    assert(node_GetNodeType(ifNode) == NODE_IF_T);
    assert(node_GetNodeType(elifNode) == NODE_IF_T);
    assert(node_GetIfNodeCondExpr(elifNode) != NULL);

    ifNode->elifNode = elifNode;
}

/**
 * @brief Задаём условному узлу альтернативный else узел
 *
 * @param [in,out] ifNode    Условный узел
 * @param [in]     elseNode  Узел альтернативы
 */
void node_SetIfNodeElse(Node_t * ifNode, Node_t * elseNode)
{
    assert(node_GetNodeType(ifNode) == NODE_IF_T);
    assert(node_GetNodeType(elseNode) == NODE_IF_T);
    assert(node_GetIfNodeCondExpr(elseNode) == NULL);

    ifNode->elifNode = elseNode;
}

/**
 * @brief node_GetIfNodeCondExpr
 *
 * @param [in] ifNode  Узел, параметр которого получаем
 *
 * @returns Узел условного узла
 */
Node_t * node_GetIfNodeCondExpr(const Node_t * ifNode)
{
    assert(node_GetNodeType(ifNode) == NODE_IF_T);

    return ifNode->condExpr;
}

/**
 * @brief Получаем узел истинной ветви исполнения
 *
 * @param [in] ifNode  Узел, параметр которого получаем
 *
 * @returns Узел истинной ветви исполнения
 */
Node_t * node_GetIfNodeBody(const Node_t * ifNode)
{
    assert(node_GetNodeType(ifNode) == NODE_IF_T);

    return ifNode->ifBodyNode;
}

/**
 * @brief Получаем узел альтернативной ветви исполнения
 *
 * @param [in] ifNode  Узел, параметр которого получаем
 *
 * @returns Узел альтернативной ветви исполнения
 */
Node_t * node_GetIfNodeAlternate(const Node_t * ifNode)
{
    assert(node_GetNodeType(ifNode) == NODE_IF_T);

    return ifNode->elifNode;
}

/**
 * @brief Получаем последний узел цепочки if-elif-else
 *
 * @param [in] ifNode  Узел, параметр которого получаем
 *
 * @returns Последний узел цепочки if-elif-else
 */
Node_t * node_GetIfNodeLastElif(const Node_t * ifNode)
{
    Node_t * node = (Node_t *)ifNode;

    while( node_GetIfNodeAlternate(node) != NULL )
    {
        node = node_GetIfNodeAlternate(node);
    }

    return node;
}


/**
 * @brief Создаём узел цикла
 *
 * @param [in] initNode      Узел инициализации цикла
 * @param [in] condNode      Узел условия выхода из цикла
 * @param [in] iterNode      Узел действия при переходе на итерацию
 * @param [in] loopBodyNode  Узел тела цикла
 *
 * @returns Узел цикла
 */
Node_t * node_CreateForNode(Node_t * initNode,
                            Node_t * condNode,
                            Node_t * iterNode,
                            Node_t * loopBodyNode)
{
    if( initNode != NULL )
    {
        assert( node_GetNodeType(initNode) == NODE_EXPRESSION_T ||
                node_GetNodeType(initNode) == NODE_OBJECT_T);
    }

    if( condNode != NULL )
    {
        assert( node_GetNodeType(condNode) == NODE_EXPRESSION_T);
    }

    if( iterNode != NULL )
    {
        assert( node_GetNodeType(iterNode) == NODE_EXPRESSION_T);
    }

    Node_t  * node;

    node = node_CreateNode(NODE_FOR_T);

    node->initNode      = initNode;
    node->condNode      = condNode;
    node->iterNode      = iterNode;
    node->loopBodyNode  = loopBodyNode;

    return node;
}

/**
 * @brief Получаем узел инициализации цикла
 *
 * @param [in] forNode  Узел, параметр которого получаем
 *
 * @returns Узел инициализации цикла
 */
Node_t * node_GetForNodeInit(const Node_t * forNode)
{
    assert( node_GetNodeType(forNode) == NODE_FOR_T);

    return forNode->initNode;
}

/**
 * @brief  Получаем узел услвия выхода цикла
 *
 * @param [in] forNode  Узел, параметр которого получаем
 *
 * @returns Узел услвия выхода цикла
 */
Node_t * node_GetForNodeCond(const Node_t * forNode)
{
    assert( node_GetNodeType(forNode) == NODE_FOR_T);

    return forNode->condNode;
}

/**
 * @brief  Получаем узел действия при переходе на итерацию
 *
 * @param [in] forNode  Узел, параметр которого получаем
 *
 * @returns Узел действия при переходе на итерацию
 */
Node_t * node_GetForNodeIter(const Node_t * forNode)
{
    assert( node_GetNodeType(forNode) == NODE_FOR_T);

    return forNode->iterNode;
}

/**
 * @brief  Получаем узел тела цикла
 *
 * @param [in] forNode  Узел, параметр которого получаем
 *
 * @returns Узел тела цикла
 */
Node_t * node_GetForNodeBody(const Node_t * forNode)
{
    assert( node_GetNodeType(forNode) == NODE_FOR_T);

    return forNode->loopBodyNode;
}


/**
 * @brief Создаём узел перехода на следующую итерацию
 *
 * @returns Узел перехода на следующую итерацию
 */
Node_t * node_CreateContinueNode(void)
{
    Node_t  * node;

    node = node_CreateNode(NODE_CONTINUE_T);

    return node;
}


/**
 * @brief Создаём узел выхода из цикла
 *
 * @returns Узел выхода из цикла
 */
Node_t * node_CreateBreakNode(void)
{
    Node_t  * node;

    node = node_CreateNode(NODE_BREAK_T);

    return node;
}


/**
 * @brief Создаём узел возврата
 *
 * @brief [in] retExpr  Узел с выражением возврата из функции
 *
 * @returns Узел выхода из цикла
 */
Node_t * node_CreateReturnNode(Node_t * retExpr)
{
    if( retExpr != NULL )
    {
        /* Если возвращаем void, то выражения не будет */
        assert(node_GetNodeType(retExpr) == NODE_LITERAL_T ||
               node_GetNodeType(retExpr) == NODE_OBJECT_T ||
               node_GetNodeType(retExpr) == NODE_EXPRESSION_T);
    }

    Node_t  * node;

    node = node_CreateNode(NODE_RETURN_T);

    node->retExpr = retExpr;

    return node;
}

/**
 * @brief Получаем узел возвращаемого выражения
 *
 * @param [in] Узел, параметр которого получаем
 *
 * @returns Узел возвращаемого выражения
 */
Node_t * node_GetReturnNodeExpr(const Node_t * retNode)
{
    assert( node_GetNodeType(retNode) == NODE_RETURN_T);

    return retNode->retExpr;
}


/**
 * @brief Создаём узел удаления массива
 *
 * @param [in] objNode  Узел объекта, который удаляем
 *
 * @returns Узел удаления массива
 */
Node_t * node_CreateDeleteNode(Node_t * delObj)
{
    Node_t  * node;

    node = node_CreateNode(NODE_DELETE_T);

    node->delObj = delObj;

    return node;
}

/**
 * @brief Получаем узел удаляемого объекта
 *
 * @param [in] deleteNode  Узел, параметр которого получаем
 *
 * @returns Узел удаляемого объекта
 */
Node_t * node_GetDeleteNodeObj(const Node_t * deleteNode)
{
    assert( node_GetNodeType(deleteNode) == NODE_DELETE_T);

    return deleteNode->delObj;
}
