/**
 * @file x-node.h
 * @brief Описание узла дерева разбора
 * @author Маркин Алексей (alexanius.ru)
 */

#ifndef NODE_H
#define NODE_H

#include <stdbool.h>

/**
 * @brief Тип выражения в узле
 */
typedef enum
{
    EXPR_ASSIGN_T,  /*!< Присваивание */
    EXPR_ADD_T,     /*!< Сложение */
    EXPR_SUB_T,     /*!< Вычитание */
    EXPR_MUL_T,     /*!< Умножение */
    EXPR_DIV_T,     /*!< Деление */
    EXPR_GRT_T,     /*!< Отношение больше */
    EXPR_LES_T,     /*!< Отношение меньше */
    EXPR_GEQ_T,     /*!< Отношение больше или равно */
    EXPR_LEQ_T,     /*!< Отношение меньше или равно */
    EXPR_EQ_T,      /*!< Отношение равно */
    EXPR_NEQ_T,     /*!< Отношение не равно */
    EXPR_LOR_T,     /*!< Логическое или */
    EXPR_LAND_T,    /*!< Логическое и */
    EXPR_LNOT_T,    /*!< Логическое не */
    EXPR_UMIN_T,    /*!< Унарный минус */
    EXPR_DECOMP_T,  /*!< Декомпозиция массива */
    EXPR_CALL_T,    /*!< Вызов процедуры */
    EXPR_END
} ExprType_t;

/**
 * @brief Тип узла в дереве разбора
 */
typedef enum
{
    NODE_MODULE_T,          /*!< Узел, содержащий модуль */
    NODE_FUNCTION_T,        /*!< Узел, содержащий функцию */
    NODE_OBJECT_T,          /*!< Узел, содержащий объект */
    NODE_EXPRESSION_T,      /*!< Узел, содержащий выражение */
    NODE_LITERAL_T,         /*!< Узел, содержащий литерал */
    NODE_IF_T,              /*!< Узел условной конструкции */
    NODE_FOR_T,             /*!< Узел цикла */
    NODE_CONTINUE_T,        /*!< Узел перехода на следующую итерацию */
    NODE_BREAK_T,           /*!< Узел выхода из цикла */
    NODE_RETURN_T,          /*!< Узел возврата */
    NODE_DELETE_T,          /*!< Узел удаления массива */
    NODE_END
} NodeType_t;

/**
 * @brief Узел дерева разбора программы
 *
 * В узле содержится информация о токене а также взаимосвязь с прочими узлами
 */
typedef struct Node_s
{
    NodeType_t      type;           /*!< Тип узла */
    struct Node_s   * nextNode;     /*!< Следующий узел */

    /* Значения, характерные для разных типов узлов */
    union
    {
        char        * moduleName;   /*!< Имя модуля */
        char        * funcName;     /*!< Имя функции */
        char        * objName;      /*!< Имя объекта */
        char        * arrName;      /*!< Имя массива */
        ExprType_t  exprType;       /*!< Тип выражения */

        /* Лексема константы - временное решение до тех пор
         * пока не сделаем систему типов */
        char        * literalLexeme;  /*!< Лексема знаения константы */
    };

    /* Здесь располагаются указатели на различные дочерние узлы */
    union
    {
        /* Поля модуля */
        struct Node_s       * firstFunc;    /*!< Первая функция модуля */

        /* Поля функции */
        struct
        {
            struct Node_s   * firstArg;     /*!< Объявление первого аргумента */
            struct Node_s   * funcBody;     /*!< Тело функции */
            struct Node_s   * retTypeNode;  /*!< Узел возвращаемого значения */
        };

        /* Поля выражения */
        struct
        {
            struct Node_s   * leftArg;      /*!< Левый аргумент выражения */
            struct Node_s   * rightArg;     /*!< Правый аргумент выражения */
        };

        /* Поля доступа к массиву */
        struct
        {
            struct Node_s   * indexExpr;    /*!< Выражение вычисления индекса */
            struct Node_s   * elemNode;     /*!< Следующее измерение массива */
        };

        /* Поля условных конструкций if-elif-else */
        struct
        {
            struct Node_s   * condExpr;     /*!< Выражение вычисления условия */
            struct Node_s   * ifBodyNode;   /*!< Тело условия (по истинному значению) */
            struct Node_s   * elifNode;     /*!< Узел альтернативного ветвления */
        };

        /* Поля цикловых конструкций for */
        struct
        {
            struct Node_s   * initNode;     /*!< Узел инициализирующего выражения */
            struct Node_s   * condNode;     /*!< Узел условия завершения цикла */
            struct Node_s   * iterNode;     /*!< Узел действия при переходе на итерацию */
            struct Node_s   * loopBodyNode; /*!< Узел тела цикла */
        };

        /* Поля возврата из функции */
        struct Node_s       * retExpr;      /*!< Выражение возврата из функции */

        /* Поля удаления массива */
        struct Node_s       * delObj;       /*!< Узел удаляемого объекта */
    };

    bool    isDecl;                 /*!< Флаг того что в узле только объявление */
} Node_t;

/* Общие функции для всех типов узлов */
void node_DeleteNode(Node_t * node);
void node_PrintNode(const Node_t * node);
Node_t * node_GetNextNode(const Node_t * node);
Node_t * node_GetLastNode(const Node_t * node);
void node_SetNextNode(Node_t * node, Node_t * nextNode);
bool node_IsDeclaration(const Node_t * node);
void node_SetNodeDeclaration(Node_t * node, bool   isDecl);

/* Работа с узлами модуля */
Node_t * node_CreateModuleNode(const char * moduleName);
const char * node_GetModuleName(const Node_t * node);
void node_SetModuleFirstFunc(Node_t * moduleNode,
                             Node_t * funcNode);

/* Работа с узлами функции */
Node_t * node_CreateFunctionNode(const char * functionName);
const char * node_GetFunctionName(const Node_t * funcNode);
void node_SetFunctionFirstArg(Node_t * funcNode, Node_t * objNode);
Node_t * node_GetFunctionFirstArg(const Node_t * funcNode);
void node_SetFunctionRetNode(Node_t * funcNode, Node_t * retNode);
Node_t * node_GetFunctionRetNode(const Node_t * funcNode);
void node_SetFunctionBody(Node_t * funcNode, Node_t * node);
Node_t * node_GetFunctionBody(const Node_t * funcNode);

/* Работа с узлами объекта */
Node_t * node_CreateObjectNode(const char * objectName);
const char * node_GetObjectName(const Node_t * objNode);

/* Работа с узлами выражения */
Node_t * node_CreateExpressionNode(ExprType_t   exprType,
                                   Node_t       * leftNode,
                                   Node_t       * rightNode);
Node_t * node_GetExprNodeLeftNode(const Node_t * exprNode);
Node_t * node_GetExprNodeRightNode(const Node_t * exprNode);
void node_SetExprNodeDecompTarg(Node_t * decompNode,
                                Node_t * targNode);
void node_SetExprNodeDecompObj(Node_t * decompNode,
                               Node_t * objNode);

/* Работа с узлами литералами */
Node_t * node_CreateLiteralNode(const char * literalLexeme);
const char * node_GetLiteralLexeme(const Node_t * literalNode);

/* Работа с узлами условия */
Node_t * node_CreateIfNode(Node_t * condExpr, Node_t * ifBodyNode);
void node_SetIfNodeElif(Node_t * ifNode, Node_t * elifNode);
void node_SetIfNodeElse(Node_t * ifNode, Node_t * elseNode);
Node_t * node_GetIfNodeCondExpr(const Node_t * ifNode);
Node_t * node_GetIfNodeBody(const Node_t * ifNode);
Node_t * node_GetIfNodeAlternate(const Node_t * ifNode);
Node_t * node_GetIfNodeLastElif(const Node_t * ifNode);

/* Работа с узлами цикла */
Node_t * node_CreateForNode(Node_t * initNode,
                            Node_t * condNode,
                            Node_t * iterNode,
                            Node_t * loopBodyNode);
Node_t * node_GetForNodeInit(const Node_t * forNode);
Node_t * node_GetForNodeCond(const Node_t * forNode);
Node_t * node_GetForNodeIter(const Node_t * forNode);
Node_t * node_GetForNodeBody(const Node_t * forNode);

/* Работа с узлами перехода на следующую итерацию */
Node_t * node_CreateContinueNode(void);

/* Работа с узлами выхода из цикла */
Node_t * node_CreateBreakNode(void);

/* Работа с узлами возврата */
Node_t * node_CreateReturnNode(Node_t * retExpr);
Node_t * node_GetReturnNodeExpr(const Node_t * retNode);

/* Работа с узлами удаления массива */
Node_t * node_CreateDeleteNode(Node_t * delObj);
Node_t * node_GetDeleteNodeObj(const Node_t * deleteNode);


#endif /* NODE_H */
